"""
Draw a random number from a beta dirstribution
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from collections import OrderedDict
import random
import numpy as np
from pynsn.dot_array.visual_features import VisualFeatures

try:
    from math import log2
except:
    from math import log

    log2 = lambda x: log(x, 2)


def is_base_string(s):
    return isinstance(s, (str, bytes))

def is_unicode_string(s):
    return isinstance(s, str)

def is_byte_string(s):
    return isinstance(s, bytes)

# randomizing
random.seed()

def random_beta(size, number_range, mean, std):
    """Draw from beta distribution defined by the
    number_range [a,b] and mean and standard distribution

    Resulting distribution has the defined mean and std

    for calculated shape parameters [alpha, beta] see `shape_parameter_beta`

    Parameter:
    ----------
    number_range : tuple (numeric, numeric)
        the range of the distribution
    mean: numeric
    std: numeric TODO

    Note:
    -----
    Depending on the position of the mean in number range the
    distribution is left or right skewed.

    """


    if std is None or number_range is None or std == 0:
        return np.array([mean]*size)

    alpha, beta = shape_parameter_beta(number_range=number_range,
                                       mean=mean,
                                       std=std)

    # NOTE: do not use np.random.beta, because it produces identical numbers for
    # different threads:
    dist = np.array([random.betavariate(alpha=alpha, beta=beta) \
                     for _ in range(size)])
    dist = (dist - np.mean(dist)) / np.std(dist) # z values
    return dist*std + mean

def shape_parameter_beta(number_range, mean, std):
    """Returns alpha (p) & beta (q) parameter for the beta distribution
    http://www.itl.nist.gov/div898/handbook/eda/section3/eda366h.htm

    Parameter
    ---------
    number_range : tuple (numeric, numeric)
        the range of the distribution
    mean : numeric
        the distribution mean
    std : numeric
         the distribution standard deviation

    Returns
    -------
    parameter: tuple
        shape parameter (alpha, beta) of the distribution

    """

    if mean <= number_range[0] or mean >= number_range[1] or \
            number_range[0] >= number_range[1]:
        raise RuntimeError("Mean has to be inside the defined number range")
    f = float(number_range[1] - number_range[0])
    mean = (mean - number_range[0]) / f
    std = (std) / f
    x = (mean * (1 - mean) / std ** 2 - 1)
    return (x * mean, (1 - mean) * x)


def join_dict_list(list_of_dicts):
    """make a dictionary of lists from a list of dictionaries"""
    rtn = OrderedDict()
    for d in list_of_dicts:
        for k, v in d.items():
            if k in rtn:
                rtn[k].append(v)
            else:
                rtn[k] = [v]
    return rtn


def dict_to_csv(dictionary, variable_names=False, dict_of_lists=False):
    d = OrderedDict(dictionary.items())
    rtn = ""
    if variable_names:
        rtn += ",".join(d.keys()) + "\n"

    if dict_of_lists:
        feat_np = np.array(list(d.values())).T  # list is requires in PY3
        for x in feat_np:
            rtn += ", ".join(map(lambda s: str(s), x)) + "\n"
    else:
        rtn += ",".join(map(lambda s: str(s), d.values())) + "\n"

    return rtn

def numpy_vector(x):
    """helper function:
    make an numpy vector from any element (list, arrays, and single data (str, numeric))
    """

    x = np.array(x)
    if x.ndim == 1:
        return x
    elif x.ndim == 0:
        return x.reshape(1)  # if one element only, make a array with one element
    else:
        return x.flatten()


def is_all_larger(vector, standard=0):
    return sum(map(lambda x: x > standard, vector))==len(vector)

def is_all_smaller(vector, standard=0):
    return sum(map(lambda x: x < standard, vector))==len(vector)

def is_all_equal(vector):
    # returns true if all elements are equal
    return sum(map(lambda x: x==vector[0], vector))==len(vector)



def check_feature_list(feature_list):
    """helper function
    raises TypeError or Runtime errors if checks fail
    * type check
    * dependency check
    """

    size_occured = ""
    space_occured = ""
    error = "Incompatible properties to match: {} & {}"

    if not isinstance(feature_list, (tuple, list)):
        feature_list = [feature_list]

    for x in feature_list:
        if x not in VisualFeatures.ALL_FEATURES:
            raise TypeError("Parameter is not a continuous feature or a " + \
                            "list of continuous properties")
            # continious property or visual feature

        if x in VisualFeatures.SIZE_FEATURES:
            if len(size_occured)>0:
                raise RuntimeError(error.format(x, size_occured))
            else:
                size_occured = x

        if x in VisualFeatures.SPACE_FEATURES:
            if len(space_occured)>0:
                raise RuntimeError(error.format(x, space_occured))
            else:
                space_occured = x





__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import math
from ..lib.coordinate2D import Coordinate2D


class Dot(Coordinate2D):

    def __init__(self, x=0, y=0, diameter=1, attribute=None):
        """Initialize a point

        Handles polar and cartesian representation (optimised processing, i.e.,
        conversions between coordinates systems will be done only once if needed)

        Parameters
        ----------
        x : numeric (default=0)
        y : numeric (default=0)
        diameter : numeric (default=1)
        attribute : attribute (string)
        """

        Coordinate2D.__init__(self, x=x, y=y)
        self.diameter = diameter
        if attribute is not None and not isinstance(attribute, str):
            raise TypeError("attributes must be a string or None, not {}".format(type(attribute).__name__))
        self.attribute = attribute

    def distance(self, d):
        """Return Euclidean distance to the dot d. The function takes the
        diameter of the points into account.

        Parameters
        ----------
        d : Dot

        Returns
        -------
        distance : float

        """

        return Coordinate2D.distance(self, d) - \
               ((self.diameter + d.diameter) / 2.0)

    @property
    def area(self):
        return math.pi * (self.diameter ** 2) / 4.0

    @property
    def perimeter(self):
        return math.pi * self.diameter


class Rectangle(Coordinate2D):

    def __init__(self, center_x=0, center_y=0, width=0, height=0, attribute=None):
        """Initialize a point

        Handles polar and cartesian representation (optimised processing, i.e.,
        conversions between coordinates systems will be done only once if needed)

        Parameters
        ----------
        x : numeric (default=0)
        y : numeric (default=0)
        width : numeric (default=1)
        height : numeric (default=1)
        attribute : attribute (string)
        """

        Coordinate2D.__init__(self, x=center_x, y=center_y)
        if attribute is not None and not isinstance(attribute, str):
            raise TypeError("attributes must be a string or None, not {}".format(type(attribute).__name__))
        self.attribute = attribute
        self.height = height
        self.width = width

    @property
    def left(self):
        return self.x - 0.5 * self.width

    @property
    def top(self):
        return self.y + 0.5 * self.height

    @property
    def right(self):
        return self.x + 0.5 * self.width

    @property
    def bottom(self):
        return self.y - 0.5 * self.height

    def is_point_inside_rect(self, xy):
        return (self.left <= xy[0] <= self.right and
                self.top <= xy[1] <= self.bottom)

    def overlaps_with(self, rect):
        d = self.xy_distances(rect)
        return not(d[0]>0 or d[1]>0)

    @property
    def area(self):
        return self.width * self.height

    @property
    def perimeter(self):
        return 2 * (self.width + self.height)

    def xy_distances(self, other):
        """return distances on both axes between rectangles. 0 indicates
        overlap off edges along that dimension.
        """

        assert isinstance(other, Rectangle)
        #overlaps in x or y
        center_dist = abs(self.x - other.x), abs(self.y - other.y)
        max_overlap_dist = (self.width + other.width)/2, (self.height + other.height)/2
        if  center_dist[0] <= max_overlap_dist[0]:
            dx = 0
        else:
            dx = center_dist[0] - max_overlap_dist[0]

        if  center_dist[1] <= max_overlap_dist[1]:
            dy = 0
        else:
            dy = center_dist[1] - max_overlap_dist[1]

        return dx, dy


    def distance(self, other):
        """Return Euclidean distance to the dot d. The function takes the
        size of the rectangles into account.

        Parameters
        ----------
        d : Dot

        Returns
        -------
        distance : float

        """

        dx, dy = self.xy_distances(other=other)
        return math.hypot(dx, dy)




"""
Draw a random number from a beta dirstribution
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import numpy as np
from scipy import spatial
from ..lib.geometry import cartesian2polar, polar2cartesian

class ConvexHull(object):
    """convenient wrapper class for calculation of convex hulls
    """

    def __init__(self, xy):
        self._xy = xy
        self.scipy_convex_hull = spatial.ConvexHull(self._xy)

    @property
    def indices(self):
        return self.scipy_convex_hull.vertices

    @property
    def xy(self):
        return self._xy[self.indices, :]


class ConvexHullDots(ConvexHull):
    """convenient wrapper class for calculation of convex hulls
    """

    def __init__(self, xy, diameters):
        ConvexHull.__init__(self, xy=xy)
        self._diameters = diameters
        self._full_xy = None
        self._full_area = None

    @property
    def full_xy(self):
        """this convex_hull takes into account the dot diameter"""
        if self._full_xy is None:
            idx = self.scipy_convex_hull.vertices

            minmax = np.array((np.min(self._xy, axis=0), np.max(self._xy, axis=0)))
            center = np.reshape(minmax[1, :] - np.diff(minmax, axis=0) / 2, 2)  # center outer positions

            polar_centered = cartesian2polar(self._xy[idx, :] - center)
            polar_centered[:, 0] = polar_centered[:, 0] + (self._diameters[idx] / 2)
            self._full_xy = polar2cartesian(polar_centered) + center

        return self._full_xy

    @property
    def full_field_area(self):
        if self._full_area is None:
            self._full_area = spatial.ConvexHull(self.full_xy).volume

        return self._full_area



__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

from PIL import Image as _Image
from PIL import ImageDraw as _ImageDraw
import numpy as _np
from ..lib import colour as _colour

def create(dot_array, colours, antialiasing=True,
           gabor_filter=None):
    # ImageParameter
    """use PIL colours (see PIL.ImageColor.colormap)

    returns pil image

    antialiasing: Ture or integer

    gabor_filter: from PIL.ImageFilter
    default_dot_colour: if colour is undefined in dot_array
    """

    if isinstance(antialiasing, bool):
        if antialiasing:  # (not if 1)
            aa = 2  # AA default
        else:
            aa = 1
    else:
        try:
            aa = int(antialiasing)
        except:
            aa = 1

    if not isinstance(colours, _colour.ImageColours):
        raise ValueError("Colours must be a ImageColours instance")

    image_size = int(round(dot_array.target_array_radius * 2)) * aa
    img = _Image.new("RGBA", (image_size, image_size),
                    color=colours.background.colour)

    if colours.target_area.colour is not None:
        _draw_dot(img, xy=_convert_pos(_np.zeros(2), image_size),
                  diameter=image_size,
                  colour=colours.target_area.colour)

    dot_array = dot_array.copy()
    dot_array.round(decimals=0)

    # draw dots
    for xy, d, att in zip(_convert_pos(dot_array.xy * aa, image_size),
                        dot_array.diameters * aa,
                        dot_array.attributes):
        if att is None:
            c = colours.default_dot_colour
        else:
            try:
                c = _colour.Colour(att)
            except:
                c = colours.default_dot_colour
        _draw_dot(img, xy=xy, diameter=d, colour=c.colour)

    if colours.field_area.colour is not None:
        # plot convey hull
        _draw_convex_hull(img=img,
                          convex_hull=_convert_pos(
                              dot_array.features.convex_hull.xy * aa, image_size),
                          convex_hull_colour=colours.field_area.colour)

    if colours.field_area_outer.colour is not None:
        # plot convey hull
        _draw_convex_hull(img=img,
                          convex_hull=_convert_pos(
                              dot_array.features.convex_hull.full_xy * aa,
                              image_size),
                          convex_hull_colour=colours.field_area_outer.colour)

    if colours.center_of_mass.colour is not None:
        _draw_dot(img, xy=_convert_pos(dot_array.center_of_mass * aa, image_size),
                  diameter=10 * aa, colour=colours.center_of_mass.colour)

    if colours.center_of_outer_positions.colour is not None:
        _draw_dot(img, xy=_convert_pos(dot_array.center_of_outer_positions * aa, image_size),
                  diameter=10 * aa, colour=colours.center_of_outer_positions.colour)

    if aa != 1:
        image_size = int(image_size / aa)
        img = img.resize((image_size, image_size), _Image.LANCZOS)

    if gabor_filter is not None:
        try:
            img = img.filter(gabor_filter)
        except:
            raise RuntimeError("Can't apply gabor_filter {}".format(gabor_filter))

    return img


def _convert_pos(xy, image_size):
    """convert dot pos to pil image coordinates"""
    return (xy * [1, -1]) + image_size // 2


def _draw_dot(img, xy, diameter, colour, picture=None):
    # draw a dot on an image

    r = diameter // 2
    if picture is not None:
        pict = _Image.open(picture, "r")
        img.paste(pict, (xy[0] - r, xy[1] - r))
    else:
        _ImageDraw.Draw(img).ellipse((xy[0] - r, xy[1] - r, xy[0] + r,
                                          xy[1] + r), fill=colour)


def _draw_convex_hull(img, convex_hull, convex_hull_colour):
    # plot convey hull

    hull = _np.append(convex_hull, [convex_hull[0]], axis=0)
    last = None
    draw = _ImageDraw.Draw(img)
    for p in hull:
        if last is not None:
            draw.line(_np.append(last, p).tolist(),
                      width=2,
                      fill=convex_hull_colour)
        last = p



"""
NonSyNum package

Creating Non-Symbolic Number Displays
"""

__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'
__version__ = '0.9.5-2'

from sys import version_info as _python_version_info
if not(_python_version_info[0] >= 3 and _python_version_info[1] >= 5):
    raise RuntimeError("PyNsN {0} ".format(__version__) +
                      "is not compatible with Python {0}.{1}. ".format(
                                                    _python_version_info[0],
                                                    _python_version_info[1]) +
                      "Please use Python 3.5+.")


from .dot_array.dot_array import DotArray
from .dot_array.shape import Dot, Rectangle
from .dot_array.visual_features import VisualFeatures
from .dot_array.match import FeatureMatcher # just for setting ITERATIVE_CONVEX_HULL_MODIFICATION and TAKE_RANDOM_DOT_FROM_CONVEXHULL toDO maybe another solution possible
from .dot_array import random_dot_array
from .lib.colour import Colour, ImageColours

#Further modules
# gui, pygame_surface, expyriment_stimulus,  dot_array_sequence, dot_array_archive




#!/usr/bin/python

"""
"""
try:
    from PyQt5.QtWidgets import QApplication as _QApplication
except:
    raise ModuleNotFoundError("Running the PyNSN GUI requires the "
                              "installation 'PyQt5'")

import sys as _sys

from .qt.gui_main_window import GUIMainWindow as _GUIMainWindow


def start():
    app = _QApplication(_sys.argv)
    ex = _GUIMainWindow()
    ex.show()
    _sys.exit(app.exec_())



#!/usr/bin/env python3
"""
Installer
"""

from setuptools import setup
import codecs
import os
from sys import version_info as _vi

package_name = "pynsn"

install_requires = ["numpy>=1.6",
                    "scipy>=1.0",
                    "Pillow>=5.0",
                    'PyQt5>=5.14',
                    "svgwrite>=1.4",
                    "pandas>=1.3"]

extras_require = {
    'pygame':             ["pygame>=1.9"],
    'expyriment':         ["expyriment>=0.9"],
}

# FIXME Docu extra requirements, make some packages suggested

entry_points = {'console_scripts': ['pynsn=pynsn.gui:start']}

packages = [package_name]
for subp in ["dot_array","image", "lib", "qt", "database", "sequence"]:
    packages.append("{}.{}".format(package_name, subp))

if _vi.major< 1:
    raise RuntimeError("{0} requires Python 3 or larger.".format(package_name))

def readme():
    directory = os.path.dirname(os.path.join(
        os.getcwd(), __file__, ))
    with codecs.open(
        os.path.join(directory, "README.md"),
        encoding="utf8",
        mode="r",
        errors="replace",
        ) as file:
        return file.read()

def get_version(package):
    """Get version number"""

    with open(os.path.join(package, "__init__.py")) as f:
        for line in f:
            if line.startswith("__version__"):
                return line.split("'")[1]
    return "None"


if __name__ == '__main__':
    setup(
        name = package_name,
        version=get_version(package_name),
        description='Creating Non-Symbolic Number Displays',
        author='Oliver Lindemann',
        author_email='lindemann@cognitive-psychology.eu',
        license='GNU GPLv3',
        url='https://github.com/lindemann09/PyNSN',
        packages=packages,
        include_package_data=True,
        setup_requires=[],
        install_requires=install_requires,
        entry_points=entry_points,
        extras_require=extras_require,
        keywords = "", #ToDo
        classifiers=[
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering"
        ],
        long_description=readme(),
        long_description_content_type='text/markdown'
    )

    try:
        import PyQt5
    except:
        print("Please note:\n"
          "Running the PyNSN GUI requires the installation 'PyQt5' ")



import random
import numpy as np
from ..lib import misc, geometry
from .visual_features import VisualFeatures

_DEFAULT_SPACING_PRECISION = 0.0001
_DEFAULT_MATCH_FA2TA_RATIO = 0.5


class FeatureMatcher(object):
    # some parameter for matching field array
    ITERATIVE_CONVEX_HULL_MODIFICATION = False  # matching convexhull TODO DOCU
    # and matching log spacing
    TAKE_RANDOM_DOT_FROM_CONVEXHULL = True  # TODO DOCU, maybe method for modification, set via GUI


    def __init__(self, dot_array):
        self.da = dot_array

    def item_diameter(self, value):
        # changes diameter

        scale = value / self.da.features.mean_item_diameter
        self.da._diameters = self.da._diameters * scale
        self.da.features.reset()

    def total_surface_area(self, value):
        # changes diameter
        a_scale = (value / self.da.features.total_surface_area)
        self.da._diameters = np.sqrt(
            self.da.surface_areas * a_scale) * 2 / np.sqrt(
            np.pi)  # d=sqrt(4a/pi) = sqrt(a)*2/sqrt(pi)
        self.da.features.reset()


    def field_area(self, value, precision=None, use_scaling_only=False):
        """changes the convex hull area to a desired size with certain precision

        uses scaling radial positions if field area has to be increased
        uses replacement of outer points (and later re-scaling)

        iterative method can takes some time.
        """

        # PROCEDURE
        #
        # increasing field area:
        #   * merely scales polar coordinates of Dots
        #   * uses __scale_field_area
        #
        # decreasing field area:
        #   a) iterative convex hull modification
        #      1. iteratively replacing outer dots to the side (random  pos.)
        #         (resulting FA is likely to small)
        #      2. increase FA by scaling to match precisely
        #         inside the field area
        #      - this methods results in very angular dot arrays, because it
        #           prefers a solution with a small number of convex hull
        #           dots
        #   b) eccentricity criterion
        #      1. determining circle with the required field area
        #      2. replacing all dots outside this circle to the inside
        #         (random pos.) (resulting FA is likely to small)
        #      3. increase FA by scaling to match precisely
        #      - this method will result is rather circulr areas


        if precision is None:
            precision = _DEFAULT_SPACING_PRECISION

        if self.da.features.field_area is None:
            return  # not defined
        elif value > self.da.features.field_area or use_scaling_only:
            # field area is currently too small or scaling is enforced
            return self.__scale_field_area(field_area=value,
                                           precision=precision)
        elif value < self.da.features.field_area:
            # field area is too large
            self.__decrease_field_area_by_replacement(
                    max_field_area=value,
                    iterative_convex_hull_modification=
                    FeatureMatcher.ITERATIVE_CONVEX_HULL_MODIFICATION)
            # ..and rescaling to avoid to compensate for possible too
            # strong decrease
            return self.__scale_field_area(field_area=value, precision=precision)
        else:
            return

    def __decrease_field_area_by_replacement(self, max_field_area,
                                            iterative_convex_hull_modification):
        """decreases filed area by recursively moving the most outer point
        to some more central free position (avoids overlapping)

        return False if not possible else True

        Note: see doc string `_match_field_area`

        """

        # centered points
        old_center = self.da.center_of_outer_positions
        self.da._xy = self.da._xy - old_center

        removed_dots = []

        if iterative_convex_hull_modification:
            while self.da.features.field_area > max_field_area:
                # remove one random outer dot and remember it
                indices = self.da.features.convex_hull.indices
                if not FeatureMatcher.TAKE_RANDOM_DOT_FROM_CONVEXHULL:
                    # most outer dot from convex hull
                    radii_outer_dots = geometry.cartesian2polar(self.da.xy[indices],
                                                                radii_only=True)
                    i = np.where(radii_outer_dots==max(radii_outer_dots))[0]
                    idx = indices[i][0]
                else:
                    # remove random
                    idx = indices[random.randint(0, len(indices)-1)]

                removed_dots.extend(self.da.get_dots(indices=[idx]))
                self.da.delete(idx)

            # add dots to free pos inside the convex hall
            for d in removed_dots:
                try:
                    d.xy = self.da.random_free_dot_position(d.diameter,
                                                   allow_overlapping=False,
                                                   prefer_inside_field_area=True)
                except:
                    raise RuntimeError("Can't find a free position while decreasing field area.\n" +\
                                        "n={}; current FA={}, max_FA={}".format(
                                            self.da.features.numerosity+1,
                                            self.da.features.field_area,
                                            max_field_area ))

                self.da.append_dot(d)

        else:
            # eccentricity criterion
            max_radius =  np.sqrt(max_field_area/np.pi) # for circle with
                                                        # required FA
            idx = np.where(geometry.cartesian2polar(self.da.xy, radii_only=True) > max_radius)[0]
            removed_dots.extend(self.da.get_dots(indices=idx))
            self.da.delete(idx)

            # add inside the circle
            min_dist = self.da.target_array_radius - max_radius + 1
            for d in removed_dots:
                try:
                    d.xy = self.da.random_free_dot_position(d.diameter,
                                            prefer_inside_field_area=False,
                                            allow_overlapping=False,
                                            min_distance_area_boarder=min_dist)
                except:
                    raise RuntimeError(
                        "Can't find a free position while decreasing field area.\n" + \
                        "n={}; current FA={}, max_FA={}".format(
                            self.da.features.numerosity + 1,
                            self.da.features.field_area,
                            max_field_area))
                self.da.append_dot(d)

        self.da._xy = self.da._xy + old_center
        self.da.features.reset()

    def __scale_field_area(self, field_area, precision):
        """change the convex hull area to a desired size by scale the polar
        positions  with certain precision

        iterative method can takes some time.

        Note: see doc string `_match_field_area`
        """

        current = self.da.features.field_area

        if current is None:
            return  # not defined

        scale = 1  # find good scale
        step = 0.1
        if field_area < current:  # current too larger
            step *= -1

        # centered points
        old_center = self.da.center_of_outer_positions
        self.da._xy = self.da._xy - old_center
        centered_polar = geometry.cartesian2polar(self.da._xy)

        # iteratively determine scale
        while abs(current - field_area) > precision:

            scale += step

            self.da._xy = geometry.polar2cartesian(centered_polar * [scale, 1])
            self.da.features.reset() # required to recalc convex hull
            current = self.da.features.field_area

            if (current < field_area and step < 0) or \
                    (current > field_area and step > 0):
                step *= -0.2  # change direction and finer grain

        self.da._xy = self.da._xy + old_center
        self.da.features.reset()


    def coverage(self, value,
                 precision=None,
                 match_FA2TA_ratio=None):

        # FIXME check drifting outwards if extra space is small and match_FA2TA_ratio=1
        # FIXME when to realign, realignment changes field_area!
        """this function changes the area and remixes to get a desired density
        precision in percent between 1 < 0

        ratio_area_convex_hull_adaptation:
            ratio of adaptation via area or via convex_hull (between 0 and 1)

        """

        print("WARNING: _match_coverage is a experimental ")
        # dens = convex_hull_area / total_surface_area
        if match_FA2TA_ratio is None:
            match_FA2TA_ratio = _DEFAULT_MATCH_FA2TA_RATIO
        elif match_FA2TA_ratio < 0 or match_FA2TA_ratio > 1:
            match_FA2TA_ratio = 0.5

        total_area_change100 = (value * self.da.features.field_area) - self.da.features.total_surface_area
        d_change_total_area = total_area_change100 * (1 - match_FA2TA_ratio)
        if abs(d_change_total_area) > 0:
            self.total_surface_area(self.da.features.total_surface_area + d_change_total_area)

        self.field_area(self.da.features.total_surface_area / value,
                        precision=precision)


    def item_perimeter(self, value):

        self.item_diameter(value / np.pi)

    def total_perimeter(self, value):
        tmp = value / (self.da.features.numerosity * np.pi)
        self.item_diameter(tmp)

    def item_surface_area(self, value):
        ta = self.da.features.numerosity * value
        self.total_surface_area(ta)

    def log_spacing(self, value, precision=None):

        logfa = 0.5 * value + 0.5 * misc.log2(
            self.da.features.numerosity)
        self.field_area(value=2 ** logfa, precision=precision)

    def log_size(self, value):
        logtsa = 0.5 * value + 0.5 * misc.log2(self.da.features.numerosity)
        self.total_surface_area(2 ** logtsa)

    def sparcity(self, value, precision = None):
        self.field_area(value= value * self.da.features.numerosity,
                        precision=precision)


    def match_feature(self, feature,
                      value = None,
                      reference_dot_array=None):
        """
        match_properties: continuous property or list of continuous properties
        several properties to be matched
        if match dot array is specified, array will be match to match_dot_array, otherwise
        the values defined in match_features is used.
        some matching requires realignement to avoid overlaps. However,
        realigment might result in a different field area. Thus, realign after
        matching for  Size parameter and realign before matching space
        parameter.

        """

        if value is None and reference_dot_array is None:
            raise ValueError("Please specify a value or a "
                             "reference_dot_array.")

        if value is not None and reference_dot_array is not None:
            raise ValueError("Please specify either a value or "
                             "reference_dot_array, not both.")

        # type check
        if feature not in VisualFeatures.ALL_FEATURES:
            raise ValueError("{} is not a visual feature.".format(feature))

        if value is None:
            value = reference_dot_array.features.get(feature)

        # Adapt
        if feature == VisualFeatures.ITEM_DIAMETER:
            self.item_diameter(value=value)

        elif feature == VisualFeatures.ITEM_PERIMETER:
            self.item_perimeter(value=value)

        elif feature == VisualFeatures.TOTAL_PERIMETER:
            self.total_perimeter(value=value)

        elif feature == VisualFeatures.ITEM_SURFACE_AREA:
            self.item_surface_area(value=value)

        elif feature == VisualFeatures.TOTAL_SURFACE_AREA:
            self.total_surface_area(value=value)

        elif feature == VisualFeatures.LOG_SIZE:
            self.log_size(value=value)

        elif feature == VisualFeatures.LOG_SPACING:
            self.log_spacing(value=value,
                             precision=_DEFAULT_SPACING_PRECISION)

        elif feature == VisualFeatures.SPARSITY:
            self.sparcity(value=value, precision=_DEFAULT_SPACING_PRECISION)

        elif feature == VisualFeatures.FIELD_AREA:
            self.field_area(value=value, precision=_DEFAULT_SPACING_PRECISION)

        elif feature == VisualFeatures.COVERAGE:
            self.coverage(value=value, precision=_DEFAULT_SPACING_PRECISION)



# calculates visual features of a dot array/ dot cloud

from collections import OrderedDict

import numpy as np
from scipy import spatial
from ..lib import misc
from .convex_hull import ConvexHullDots


class VisualFeatures(object):

    LOG_SIZE = "Log Size"
    TOTAL_SURFACE_AREA = "Total surface area"
    ITEM_DIAMETER = "Mean item diameter"
    ITEM_SURFACE_AREA = "Mean item surface area"
    ITEM_PERIMETER = "Total perimeter"
    TOTAL_PERIMETER = "Mean item perimeter"
    LOG_SPACING = "Log Spacing"
    SPARSITY = "Sparsity"
    FIELD_AREA = "Field area"
    COVERAGE = "Coverage"

    SIZE_FEATURES = (LOG_SIZE, TOTAL_SURFACE_AREA, ITEM_DIAMETER,
                     ITEM_SURFACE_AREA, ITEM_PERIMETER, TOTAL_PERIMETER)

    SPACE_FEATURES = (LOG_SPACING, SPARSITY, FIELD_AREA)

    ALL_FEATURES = SIZE_FEATURES + SPACE_FEATURES + (COVERAGE,)

    @staticmethod
    def are_dependent(featureA, featureB):
        """returns true if both features are not independent"""
        for l in [VisualFeatures.SIZE_FEATURES, VisualFeatures.SPACE_FEATURES]:
            if featureA in l and featureB in l:
                return True
        return False


    def __init__(self, dot_array):
        # dot_array or dot_cloud
        self.da = dot_array
        self._convex_hull = None

    def reset(self):
        """reset to enforce recalculation of certain parameter """
        self._convex_hull = None

    @property
    def convex_hull(self):
        if self._convex_hull is None:
            self._convex_hull = ConvexHullDots(self.da.xy, self.da.diameters)
        return self._convex_hull

    @property
    def mean_item_diameter(self):
        return np.mean(self.da.diameters)

    @property
    def total_surface_area(self):
        return np.sum(self.da.surface_areas)

    @property
    def mean_item_surface_area(self):
        return np.mean(self.da.surface_areas)

    @property
    def total_perimeter(self):
        return np.sum(self.da.perimeter)

    @property
    def mean_item_perimeter(self):
        return np.mean(self.da.perimeter)

    @property
    def field_area(self):
        return self.convex_hull.scipy_convex_hull.volume

    @property
    def numerosity(self):
        return len(self.da.xy)

    @property
    def converage(self):
        """ percent coverage in the field area. It takes thus the item size
        into account. In contrast, the sparsity is only the ratio of field
        array and numerosity

        """

        try:
            return self.total_surface_area / self.field_area
        except:
            return None

    @property
    def logSize(self):
        return misc.log2(self.total_surface_area) + misc.log2(
            self.mean_item_surface_area)

    @property
    def logSpacing(self):
        return misc.log2(self.field_area) + misc.log2(self.sparsity)

    @property
    def sparsity(self):
        return self.field_area / self.numerosity

    @property
    def field_area_full(self):  # TODO not tested
        return self.convex_hull.full_field_area

    def _get_distance_matrix(self, between_positions=False):
        """between position ignores the dot size"""
        dist = spatial.distance.cdist(self.da.xy, self.da.xy)  #
        # matrix with all distance between all points
        if not between_positions:
            # subtract dot diameter
            radii_mtx = np.ones((self.numerosity, 1)) + \
                        self.da.diameters[:, np.newaxis].T / 2
            dist -= radii_mtx  # for each row
            dist -= radii_mtx.T  # each each column
        return dist

    @property
    def expension(self):
        """ maximal distance between to points plus diameter of the two points"""

        dist = self._get_distance_matrix(between_positions=True)
        # add dot diameter
        radii_mtx = np.ones((self.numerosity, 1)) + self.da.diameters[:,
                                                    np.newaxis].T / 2
        dist += radii_mtx  # add to each row
        dist += radii_mtx.T  # add two each column
        return np.max(dist)

    #@property
    #def featureXX_coverage_target_area(self):
    #    """density takes into account the full possible target area (i.e.,
    #        image radius) """
    #    try:
    #        return np.pi * self.da.target_array_radius ** 2 / \
    #            self.total_surface_area
    #    except:
    #        return None # dot defined for DotCloud with no target_array_radius

    def get(self, feature):
        """returns a feature"""

       # Adapt
        if feature == VisualFeatures.ITEM_DIAMETER:
            return self.mean_item_diameter

        elif feature == VisualFeatures.ITEM_PERIMETER:
            return self.mean_item_perimeter

        elif feature == VisualFeatures.TOTAL_PERIMETER:
            return self.total_perimeter

        elif feature == VisualFeatures.ITEM_SURFACE_AREA:
            return self.mean_item_surface_area

        elif feature == VisualFeatures.TOTAL_SURFACE_AREA:
            return self.total_surface_area

        elif feature == VisualFeatures.LOG_SIZE:
            return self.logSize

        elif feature == VisualFeatures.LOG_SPACING:
            return self.logSpacing

        elif feature == VisualFeatures.SPARSITY:
            return self.sparsity

        elif feature == VisualFeatures.FIELD_AREA:
            return self.field_area

        elif feature == VisualFeatures.COVERAGE:
            return self.converage

        else:
            raise ValueError("{} is a unkown visual feature".format(feature))


    def get_features_dict(self):
        """ordered dictionary with the most important feature"""
        rtn = [("Hash", self.da.hash),
               ("Numerosity", self.numerosity),
               (VisualFeatures.TOTAL_SURFACE_AREA, self.total_surface_area),
               (VisualFeatures.ITEM_SURFACE_AREA, self.mean_item_surface_area),
               (VisualFeatures.ITEM_DIAMETER, self.mean_item_diameter),
               (VisualFeatures.ITEM_PERIMETER, self.mean_item_perimeter),
               (VisualFeatures.TOTAL_PERIMETER, self.total_perimeter),
               (VisualFeatures.FIELD_AREA, self.field_area),
               (VisualFeatures.SPARSITY, self.sparsity),
               (VisualFeatures.COVERAGE, self.converage),
               (VisualFeatures.LOG_SIZE, self.logSize),
               (VisualFeatures.LOG_SPACING, self.logSpacing)]
        return OrderedDict(rtn)

    def get_features_text(self, with_hash=True, extended_format=False, spacing_char="."):
        if extended_format:
            rtn = None
            for k, v in self.get_features_dict().items():
                if rtn is None:
                    if with_hash:
                        rtn = "- {}: {}\n".format(k, v)
                    else:
                        rtn = ""
                else:
                    if rtn == "":
                        name = "- " + k
                    else:
                        name = "  " + k
                    try:
                        value = "{0:.2f}\n".format(v)  # try rounding
                    except:
                        value = "{}\n".format(v)

                    rtn += name + (spacing_char * (22 - len(name))) + (" " * (14 - len(value))) + value
        else:
            if with_hash:
                rtn = "ID: {} ".format(self.da.hash)
            else:
                rtn = ""
            rtn += "N: {}, TSA: {}, ISA: {}, FA: {}, SPAR: {:.3f}, logSIZE: " \
                   "{:.2f}, logSPACE: {:.2f} COV: {:.2f}".format(
                self.numerosity,
                int(self.total_surface_area),
                int(self.mean_item_surface_area),
                int(self.field_area),
                self.sparsity,
                self.logSize,
                self.logSpacing,
                self.converage)
        return rtn






__author__ = 'Oliver Lindemann <lindemann@cognitive-psychology.eu>'

import math

class Coordinate2D(object):

    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def __add__(self, other):
        return Coordinate2D(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Coordinate2D(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        return Coordinate2D(self.x * other, self.y * other)

    def __div__(self, other):
        return Coordinate2D(self.x / other if other else self.x,
                            self.y / other if other else self.y)

    def __iadd__(self, other):
        self.x += other.x
        self.y += other.y
        return self

    def __isub__(self, other):
        self.x -= other.x
        self.y -= other.y
        return self

    def __imul__(self, other):
        self.x *= other
        self.y *= other
        return self

    def __idiv__(self, other):
        self.x /= other
        self.y /= other
        return self

    @property
    def xy(self):
        return (self.x, self.y)

    @xy.setter
    def xy(self, xy_tuple):
        self.x = xy_tuple[0]
        self.y = xy_tuple[1]

    @property
    def polar_radius(self):
        return math.hypot(self.x, self.y)

    @polar_radius.setter
    def polar_radius(self, value):
        self.polar = (value, self.polar_angle)

    @property
    def polar_angle(self):
        return math.atan2(self.y, self.x)

    @polar_angle.setter
    def polar_angle(self, value):
        self.polar = (self.polar_radius, value)

    @property
    def polar(self):
        """polar coordinate (radius, pos_angle) """
        return (self.polar_radius, self.polar_angle)

    @polar.setter
    def polar(self, rad_ang):
        """polar coordinate (radius, angle) """

        self.x = rad_ang[0] * math.cos(rad_ang[1])
        self.y = rad_ang[0] * math.sin(rad_ang[1])


    def distance(self, d):
        """Returns Euclidean distance to the another Coordinate. The function
        does not takes the size of an object into account.

        Parameters
        ----------
        d : Coordinate2D

        Returns
        -------
        distance : float

        """

        return math.hypot(self.x - d.x, self.y - d.y)
